@@ -1,4 +1,4 @@
-/* Totem Basic Plugin
+/* Totem NarrowSpace plugin scriptable
  *
  * Copyright (C) 2004 Bastien Nocera <hadess@hadess.net>
  * Copyright (C) 2002 David A. Schleef <ds@schleef.org>
@@ -18,11 +18,11 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  *
- * $Id: totemNarrowSpacePlugin.h 3408 2006-08-11 13:25:44Z chpe $
+ * $Id: totemNarrowSpacePlugin.h 4688 2007-09-10 10:27:46Z hadess $
  */
 
-#ifndef __GMP_PLAYER_H__
-#define __GMP_PLAYER_H__
+#ifndef __NARROWSPACE_PLUGIN_H__
+#define __NARROWSPACE_PLUGIN_H__
 
 #include <nsIClassInfo.h>
 
@@ -37,17 +37,46 @@ class totemScriptablePlugin : public totemINarrowSpacePlayer,
     NS_DECL_TOTEMINARROWSPACEPLAYER
     NS_DECL_NSICLASSINFO
 
+    void* operator new (size_t aSize) CPP_THROW_NEW;
+
     totemScriptablePlugin (totemPlugin *aPlugin);
 
     PRBool IsValid () { return mPlugin != nsnull; }
-    void UnsetPlugin () { mPlugin = nsnull; }
+    void SetPlugin (totemPlugin *aPlugin) { mPlugin = aPlugin; }
 
     static char *PluginDescription ();
     static void PluginMimeTypes (const totemPluginMimeEntry **, PRUint32 *);
+
+    enum PluginState {
+      eState_Complete,
+      eState_Error,
+      eState_Loading,
+      eState_Playable,
+      eState_Waiting
+    };
+
+    PRUint32 mPluginState : 3; /* enough bits for PluginState enum values */
+
   private:
     ~totemScriptablePlugin ();
 
     totemPlugin *mPlugin;
+
+    nsCString mBackgroundColour;
+    nsCString mMatrix;
+    nsCString mRectangle;
+    nsCString mMovieName;
+
+    PRInt32 mVolume;
+
+    PRUint32 mAutoPlay : 1;
+    PRUint32 mControllerVisible : 1;
+    PRUint32 mIsLooping : 1;
+    PRUint32 mKioskMode : 1;
+    PRUint32 mLoopIsPalindrome : 1;
+    PRUint32 mMute : 1;
+    PRUint32 mPlayEveryFrame : 1;
+    PRUint32 mResetPropertiesOnReload : 1;
 };
 
-#endif /* __GMP_PLAYER_H__ */
+#endif /* __NARROWSPACE_PLUGIN_H__ */
